import numpy as np
import torch.nn as nn
import scipy
import torch
from gymnasium.spaces import Box, Discrete, Space


def combined_shape(length, shape=None):
    if shape is None:
        return (length,)
    return (length, shape) if np.isscalar(shape) else (length, *shape)


def keys_as_sorted_list(dict):
    return sorted(list(dict.keys()))


def values_as_sorted_list(dict):
    return [dict[k] for k in keys_as_sorted_list(dict)]


def mlp(sizes, activation, output_activation=nn.Identity):
    layers = []
    for j in range(len(sizes) - 1):
        act = activation if j < len(sizes) - 2 else output_activation
        layers += [nn.Linear(sizes[j], sizes[j + 1]), act()]
    return nn.Sequential(*layers)


def count_vars(module):
    return sum([np.prod(p.shape) for p in module.parameters()])


def discount_cumsum(x, discount):
    """
    magic from rllab for computing discounted cumulative sums of vectors.

    input:
        vector x,
        [x0,
         x1,
         x2]

    output:
        [x0 + discount * x1 + discount^2 * x2,
         x1 + discount * x2,
         x2]
    """
    return scipy.signal.lfilter([1], [1, float(-discount)], x[::-1], axis=0)[::-1]


class Actor(nn.Module):
    def _distribution(self, obs):
        raise NotImplementedError

    def _log_prob_from_distribution(self, pi, act):
        raise NotImplementedError

    def forward(self, obs, act=None):
        pi = self._distribution(obs)
        logp_a = None
        if act is not None:
            logp_a = self._log_prob_from_distribution(pi, act)
        return pi, logp_a


class MLPActor(nn.Module):
    def __init__(self, obs_dim, act_dim, hidden_sizes, activation, act_limit) -> None:
        super().__init__()
        pi_sizes = [obs_dim] + list(hidden_sizes) + [act_dim]
        self.pi = mlp(pi_sizes, activation, nn.Tanh)
        self.act_limit = act_limit

    def forward(self, obs):
        return self.pi(obs) * self.act_limit


class MLPQFunction(nn.Module):
    def __init__(self, obs_dim, act_dim, hidden_sizes, activation) -> None:
        super().__init__()
        self.q = mlp([obs_dim + act_dim] + list(hidden_sizes) + [1], activation)

    def forward(self, obs, act):
        q = self.q(torch.cat([obs, act], dim=-1))
        return torch.squeeze(q, -1)


class MLPQActorCritic(nn.Module):
    def __init__(
        self,
        observation_space,
        action_space,
        hidden_sizes=[256, 256],
        activation=nn.ReLU,
    ):
        super().__init__()
        obs_dim = observation_space.shape
        act_dim = action_space.shape[0]
        act_limit = action_space.hight[0]

        self.pi = MLPActor(obs_dim, act_dim, hidden_sizes, activation, act_limit)
        self.q = MLPQFunction(obs_dim, act_dim, hidden_sizes, activation)

    def act(self, obs):
        with torch.no_grad():
            return self.pi(obs).numpy()


class MLPQ2ActorCritic(nn.Module):
    def __init__(
        self,
        observation_space,
        action_space,
        hidden_sizes=[256, 256],
        activation=nn.ReLU,
    ):
        super().__init__()
        obs_dim = observation_space.shape
        act_dim = action_space.shape[0]
        act_limit = action_space.hight[0]

        self.pi = MLPActor(obs_dim, act_dim, hidden_sizes, activation, act_limit)
        self.q1 = MLPQFunction(obs_dim, act_dim, hidden_sizes, activation)
        self.q2 = MLPQFunction(obs_dim, act_dim, hidden_sizes, activation)

    def act(self, obs):
        with torch.no_grad():
            return self.pi(obs).numpy()


class MLPCategoricalActor(Actor):
    def __init__(self, obs_dim, act_dim, hidden_sizes, activation):
        super().__init__()
        self.logits_net = mlp(
            sizes=[obs_dim] + list(hidden_sizes) + [act_dim], activation=activation
        )

    def _distribution(self, obs):
        logits = self.logits_net(obs)
        return torch.distributions.Categorical(logits=logits)

    def _log_prob_from_distribution(self, pi, act):
        return pi.log_prob(act)


class MLPGuassianActor(Actor):
    def __init__(self, obs_dim, act_dim, hidden_sizes, activation):
        super().__init__()
        log_std = -0.5 * np.ones(act_dim, dtype=np.float32)
        self.log_std = torch.nn.Parameter(torch.as_tensor(log_std))
        self.mu_net = mlp(
            sizes=[obs_dim] + list(hidden_sizes) + [act_dim], activation=activation
        )

    def _distribution(self, obs):
        mu = self.mu_net(obs)
        std = torch.exp(self.log_std)
        return torch.distributions.Normal(mu, std)

    def _log_prob_from_distribution(self, pi, act):
        return pi.log_prob(act).sum(axis=-1)


class MLPCritic(nn.Module):
    def __init__(self, obs_dim, hidden_sizes, activation):
        super().__init__()
        self.v_net = mlp(
            sizes=[obs_dim] + list(hidden_sizes) + [1], activation=activation
        )

    def forward(self, obs):
        return self.v_net(obs).squeeze(dim=-1)


class MlpDklActorCritic(nn.Module):
    def __init__(
        self,
        observation_space: Space,
        action_space,
        hidden_sizes=[64, 64],
        activation=nn.Tanh,
    ):
        super().__init__()

        obs_dim = observation_space.shape[0]
        if isinstance(action_space, Box):
            self.pi = MLPGuassianActor(
                obs_dim, action_space.shape[0], hidden_sizes, activation
            )
        elif isinstance(action_space, Discrete):
            self.pi = MLPCategoricalActor(
                obs_dim, action_space.n, hidden_sizes, activation
            )
        self.v = MLPCritic(obs_dim, hidden_sizes, activation)

    def step(self, obs):
        with torch.no_grad():
            pi = self.pi._distribution(obs)
            a = pi.sample()
            logp_a = self.pi._log_prob_from_distribution(pi, a)
            v = self.v(obs)
        return a.numpy(), v.numpy(), logp_a.numpy()


class MLPActorCritic(nn.Module):
    def __init__(
        self,
        observation_space: Space,
        action_space: Space,
        hidden_sizes=[64, 64],
        activation=nn.Tanh,
    ):
        super().__init__()

        obs_dim = observation_space.shape[0]
        if isinstance(action_space, Box):
            self.pi = MLPGuassianActor(
                obs_dim, action_space.shape[0], hidden_sizes, activation
            )
        elif isinstance(action_space, Discrete):
            self.pi = MLPCategoricalActor(
                obs_dim, action_space.n, hidden_sizes, activation
            )
        self.v = MLPCritic(obs_dim, hidden_sizes, activation)

    def step(self, obs):
        with torch.no_grad():
            pi = self.pi._distribution(obs)
            a = pi.sample()
            logp_a = self.pi._log_prob_from_distribution(pi, a)
            v = self.v(obs)
        return a.numpy(), v.numpy(), logp_a.numpy()

    def act(self, obs):
        return self.step(obs)[0]


class GAEBuffer(object):
    def __init__(self, obs_dim: int, act_dim: int, size: int, gamma: float, lam: float):
        self.obs_buf = np.zeros(combined_shape(size, obs_dim), dtype=np.float32)
        self.act_buf = np.zeros(combined_shape(size, act_dim), dtype=np.float32)
        self.logp_buf = np.zeros(size, dtype=np.float32)
        self.val_buf = np.zeros(size, dtype=np.float32)
        self.rew_buf = np.zeros(size, dtype=np.float32)
        self.adv_buf = np.zeros(size, dtype=np.float32)
        self.ret_buf = np.zeros(size, dtype=np.float32)
        self.gamma, self.lam = gamma, lam
        self.ptr, self.path_start_idx, self.max_size = 0, 0, size

    def store(self, obs, act, val, rew, logp):
        self.obs_buf[self.ptr] = obs
        self.act_buf[self.ptr] = act
        self.logp_buf[self.ptr] = logp
        self.val_buf[self.ptr] = val
        self.rew_buf[self.ptr] = rew
        self.ptr += 1

    def finish_path(self, last_val=0):
        path_slice = slice(self.path_start_idx, self.ptr)
        rews = np.append(self.rew_buf[path_slice], last_val)
        vals = np.append(self.val_buf[path_slice], last_val)

        # the next two lines implement GAE-Lambda advantage calculation
        deltas = rews[:-1] + self.gamma * vals[1:] - vals[:-1]
        self.adv_buf[path_slice] = discount_cumsum(deltas, self.gamma * self.lam)

        # the next line computes rewards-to-go, to be targets for the value function
        self.ret_buf[path_slice] = discount_cumsum(rews, self.gamma)[:-1]
        self.path_start_idx = self.ptr

    def get(self):
        assert self.ptr == self.max_size
        self.ptr, self.path_start_idx = 0, 0

        # advantage normalization trick
        adv_mean, adv_std = np.mean(self.adv_buf), np.std(self.adv_buf)
        self.adv_buf = (self.adv_buf - adv_mean) / adv_std

        ret = dict(
            obs=self.obs_buf,
            act=self.act_buf,
            ret=self.ret_buf,
            adv=self.adv_buf,
            logp=self.logp_buf,
        )
        return {k: torch.as_tensor(v, dtype=torch.float32) for k, v in ret.items()}



import numpy as np
import core
import gymnasium as gym
import torch
import torch.nn as nn


class GAEBuffer(object):
    def __init__(self, obs_dim, act_dim, size, info_shapes, gamma, lam):
        self.obs_buf = np.zeros(core.combined_shape(size, obs_dim), dtype=np.float32)
        self.act_buf = np.zeros(core.combined_shape(size, act_dim), dtype=np.float32)
        self.rew_buf = np.zeros(size, dtype=np.float32)
        self.val_buf = np.zeros(size, dtype=np.float32)
        self.logp_buf = np.zeros(size, dtype=np.float32)
        self.ret_buf = np.zeros(size, dtype=np.float32)
        self.adv_buf = np.zeros(size, dtype=np.float32)
        self.info_bufs = {
            k: np.zeros([size] + list(v), dtype=np.float32)
            for k, v in info_shapes.items()
        }
        self.info_sorted_keys = core.keys_as_sorted_list(self.info_bufs)
        self.gamma, self.lam = gamma, lam
        self.ptr, self.path_start_idx, self.max_size = 0, 0, size

    def store(self, obs, act, rew, val, logp, info):
        assert self.ptr < self.max_size
        self.obs_buf[self.ptr] = obs
        self.act_buf[self.ptr] = act
        self.rew_buf[self.ptr] = rew
        self.val_buf[self.ptr] = val
        self.logp_buf[self.ptr] = logp
        for i, k in enumerate(self.info_sorted_keys):
            self.info_bufs[k][self.ptr] = info[i]
        self.ptr += 1

    def finish_path(self, last_val=0):
        path_slice = slice(self.path_start_idx, self.ptr)
        rews = np.append(self.rew_buf[path_slice], last_val)
        vals = np.append(self.val_buf[path_slice], last_val)

        deltas = rews[:-1] + self.gamma * vals[1:] - vals[:-1]
        self.adv_buf[path_slice] = core.discount_cumsum(deltas, self.gamma * self.lam)
        self.ret_buf[path_slice] = core.discount_cumsum(rews, self.gamma)[:-1]

        self.path_start_idx = self.ptr

    def get(self):
        assert self.ptr == self.max_size
        self.path_start_idx, self.ptr = 0, 0

        adv_mean, adv_std = np.mean(self.adv_buf), np.std(self.adv_buf)
        self.adv_buf = (self.adv_buf - adv_mean) / adv_std

        return [
            self.obs_buf,
            self.act_buf,
            self.adv_buf,
            self.ret_buf,
            self.logp_buf,
        ] + core.values_as_sorted_list(self.info_bufs)


def trpo(
    env_fn=lambda: gym.make("CarPole-v1"),
    actor_critic=core.MLPActorCritic,
    ac_kwargs=dict(),
    seed=0,
    epochs=50,
    max_steps=5000,
    max_ep_len=1000,
    gamm=0.99,
    lam=0.97,
    delta=0.01,
    vf_lr=1e-3,
    train_v_iters=80,
    cg_iters=10,
    backtrack_iters=10,
    backtrack_coeff=0.8,
):
    env = env_fn()
    obs_dim = env.observation_space.shape
    act_dim = env.action_space.shape
    
    # Set random seeds
    np.random.seed(seed)
    torch.manual_seed(seed)
    
    # Create actor-critic module
    ac = actor_critic(env.observation_space, env.action_space, **ac_kwargs)
    buf = GAEBuffer()



from core import discount_cumsum, combined_shape, count_vars
import torch.nn as nn
from gymnasium import Env
import core
import gymnasium as gym
import numpy as np
import torch
import logging

logging.basicConfig(
    level=logging.INFO, format="%(asctime)s [%(levelname)s] %(message)s"
)
logger = logging.getLogger(__name__)


class VPGBuffer(object):
    def __init__(self, obs_dim: int, act_dim: int, size: int, gamma: float, lam: float):
        self.obs_buf = np.zeros(combined_shape(size, obs_dim), dtype=np.float32)
        self.act_buf = np.zeros(combined_shape(size, act_dim), dtype=np.float32)
        self.logp_buf = np.zeros(size, dtype=np.float32)
        self.val_buf = np.zeros(size, dtype=np.float32)
        self.rew_buf = np.zeros(size, dtype=np.float32)
        self.adv_buf = np.zeros(size, dtype=np.float32)
        self.ret_buf = np.zeros(size, dtype=np.float32)
        self.gamma, self.lam = gamma, lam
        self.ptr, self.path_start_idx, self.max_size = 0, 0, size

    def store(self, obs, act, val, rew, logp):
        self.obs_buf[self.ptr] = obs
        self.act_buf[self.ptr] = act
        self.logp_buf[self.ptr] = logp
        self.val_buf[self.ptr] = val
        self.rew_buf[self.ptr] = rew
        self.ptr += 1

    def finish_path(self, last_val=0):
        path_slice = slice(self.path_start_idx, self.ptr)
        rews = np.append(self.rew_buf[path_slice], last_val)
        vals = np.append(self.val_buf[path_slice], last_val)

        # the next two lines implement GAE-Lambda advantage calculation
        deltas = rews[:-1] + self.gamma * vals[1:] - vals[:-1]
        self.adv_buf[path_slice] = discount_cumsum(deltas, self.gamma * self.lam)

        # the next line computes rewards-to-go, to be targets for the value function
        self.ret_buf[path_slice] = discount_cumsum(rews, self.gamma)[:-1]
        self.path_start_idx = self.ptr

    def get(self):
        assert self.ptr == self.max_size
        self.ptr, self.path_start_idx = 0, 0

        # advantage normalization trick
        adv_mean, adv_std = np.mean(self.adv_buf), np.std(self.adv_buf)
        self.adv_buf = (self.adv_buf - adv_mean) / adv_std

        ret = dict(
            obs=self.obs_buf,
            act=self.act_buf,
            ret=self.ret_buf,
            adv=self.adv_buf,
            logp=self.logp_buf,
        )
        return {k: torch.as_tensor(v, dtype=torch.float32) for k, v in ret.items()}


def vpg(
    env_fn,
    actor_critic,
    ac_kwargs: dict,
    epoch: int,
    max_steps: int,
    max_ep_len: int,
    train_v_iters: int,
    pi_lr: float,
    vf_lr: float,
    gamma: float,
    lam: float,
):
    env = env_fn()
    obs_dim = env.observation_space.shape
    act_dim = env.action_space.shape
    logger.info(f"Observation space: {obs_dim}, Action space: {act_dim}")

    ac = actor_critic(env.observation_space, env.action_space, **ac_kwargs)
    buf = VPGBuffer(obs_dim, act_dim, max_steps, gamma, lam)

    var_counts = tuple(count_vars(module) for module in [ac.pi, ac.v])
    print("\nNumber of parameters: \t pi: %d, \t v: %d\n" % var_counts)
    optimizer_pi = torch.optim.Adam(ac.pi.parameters(), lr=pi_lr)
    optimizer_v = torch.optim.Adam(ac.v.parameters(), lr=vf_lr)

    def compute_loss_pi(data):
        obs, act, adv = data["obs"], data["act"], data["adv"]
        _, logp = ac.pi(obs, act)
        loss_pi = -(logp * adv).mean()
        return loss_pi

    def compute_loss_v(data):
        obs, ret = data["obs"], data["ret"]
        return ((ac.v(obs) - ret) ** 2).mean()

    def update():
        data = buf.get()

        optimizer_pi.zero_grad()
        loss_pi = compute_loss_pi(data)
        loss_pi.backward()
        optimizer_pi.step()

        for i in range(train_v_iters):
            optimizer_v.zero_grad()
            loss_v = compute_loss_v(data)
            loss_v.backward()
            optimizer_v.step()

        return loss_pi.item(), loss_v.item()

    (o, o_info), ep_ret, ep_len = env.reset(), 0, 0
    best_reward = -np.inf
    for epoch in range(epoch):
        ep_rews = 0
        episodes = 0
        for t in range(max_steps):
            a, v, logp = ac.step(torch.as_tensor(o, dtype=torch.float32))
            next_o, r, d, tructated, _ = env.step(a)
            ep_ret += r
            ep_len += 1
            ep_rews += r

            buf.store(o, a, v, r, logp)
            o = next_o

            timeout = ep_len == max_ep_len
            terminal = d or tructated or timeout

            epoch_ended = t == max_steps - 1
            if terminal or epoch_ended:
                if epoch_ended and not terminal:
                    print(
                        "Warning: trajectory cut off by epoch at %d steps." % ep_len,
                        flush=True,
                    )
                last_val = (
                    0
                    if d or tructated
                    else ac.step(torch.as_tensor(o, dtype=torch.float32))[1].item()
                )
                buf.finish_path(last_val)
                (o, o_info), ep_ret, ep_len = env.reset(), 0, 0
                episodes += 1

        loss_pi, loss_v = update()
        ep_rews /= episodes
        if ep_rews > best_reward:
            best_reward = ep_rews
            torch.save(ac.state_dict(), "vpg.pt")
        logger.info(
            f"Epoch: {epoch}, Loss pi: {loss_pi}, Loss v: {loss_v}, Reward: {ep_rews}"
        )


def vpg_test(env_fn, actor_critic, ac_kwargs, test_epochs, model_path):
    env = env_fn()
    ac = actor_critic(env.observation_space, env.action_space, **ac_kwargs)
    ac.load_state_dict(torch.load(model_path))
    ac.eval()

    for epoch in range(test_epochs):
        (o, o_info), ep_ret, ep_len = env.reset(), 0, 0
        while True:
            a, v, logp = ac.step(torch.as_tensor(o, dtype=torch.float32))
            next_o, r, d, tructated, _ = env.step(a)
            ep_ret += r
            ep_len += 1
            o = next_o
            terminal = d or tructated
            if terminal:
                print(f"Epoch: {epoch}, Reward: {ep_ret}")
                (o, o_info), ep_ret, ep_len = env.reset(), 0, 0
                break
        cmd = input("Continue? [y/n]")
        if cmd == "n":
            break


if __name__ == "__main__":
    import argparse

    parser = argparse.ArgumentParser()
    parser.add_argument("--test", action="store_true")
    args = parser.parse_args()

    if args.test:
        vpg_test(
            lambda: gym.make("LunarLander-v2", render_mode="human"),
            actor_critic=core.MLPActorCritic,
            ac_kwargs=dict(hidden_sizes=[64, 64], activation=nn.Tanh),
            test_epochs=10,
            model_path="vpg.pt",
        )
    else:
        vpg(
            lambda: gym.make("LunarLander-v2"),
            actor_critic=core.MLPActorCritic,
            ac_kwargs=dict(hidden_sizes=[64, 64], activation=nn.Tanh),
            epoch=500,
            max_steps=5000,
            max_ep_len=1000,
            train_v_iters=80,
            pi_lr=3e-4,
            vf_lr=1e-3,
            gamma=0.99,
            lam=0.97,
        )



import torch
import gymnasium as gym
import core
import numpy as np
import torch.nn as nn
import logging

logging.basicConfig(
    level=logging.INFO, format="%(asctime)s [%(levelname)s] %(message)s"
)


def ppo(
    env_fn,
    ac_constructor,
    ac_kwargs,
    epochs,
    max_steps,
    max_ep_len,
    gamma,
    lam,
    clip_ratio,
    pi_lr,
    vf_lr,
    train_pi_iters,
    train_v_iters,
    target_kl,
    logger,
    save_path,
):
    env = env_fn()
    obs_dim = env.observation_space.shape
    act_dim = env.action_space.shape
    logger.info(f"Observation space: {obs_dim}, Action space: {act_dim}")
    buf = core.GAEBuffer(obs_dim, act_dim, max_steps, gamma, lam)
    ac = ac_constructor(env.observation_space, env.action_space, **ac_kwargs)

    pi_optimizer = torch.optim.Adam(ac.pi.parameters(), lr=pi_lr)
    vf_optimizer = torch.optim.Adam(ac.v.parameters(), lr=vf_lr)

    def compute_loss_pi(data):
        obs, act, adv, logp_old = data["obs"], data["act"], data["adv"], data["logp"]

        # policy loss
        pi, logp = ac.pi(obs, act)
        ratio = torch.exp(logp - logp_old)
        clip_adv = torch.clamp(ratio, 1 - clip_ratio, 1 + clip_ratio) * adv
        loss_pi = -(torch.min(ratio * adv, clip_adv)).mean()

        # extra useful info
        approx_kl = (logp_old - logp).mean().item()
        ent = pi.entropy().mean().item()
        clipped = ratio.gt(1 + clip_ratio) | ratio.lt(1 - clip_ratio)
        clipfrac = torch.as_tensor(clipped, dtype=torch.float32).mean().item()
        pi_info = dict(kl=approx_kl, ent=ent, cf=clipfrac)
        return loss_pi, pi_info

    def compute_loss_v(data):
        obs, ret = data["obs"], data["ret"]
        return ((ac.v(obs) - ret) ** 2).mean()

    def update():
        data = buf.get()

        pi_l_old, pi_info_old = compute_loss_pi(data)
        pi_l_old = pi_l_old.item()
        v_l_old = compute_loss_v(data).item()

        for i in range(train_pi_iters):
            pi_optimizer.zero_grad()
            loss_pi, pi_info = compute_loss_pi(data)
            loss_pi.backward()
            pi_optimizer.step()

            if pi_info["kl"] > 1.5 * target_kl:
                logger.info("Early stopping at step %d due to reaching max kl." % i)
                break

        for i in range(train_v_iters):
            vf_optimizer.zero_grad()
            loss_v = compute_loss_v(data)
            loss_v.backward()
            vf_optimizer.step()

        return loss_pi.item(), loss_v.item()

    o, ep_ret, ep_len = env.reset()[0], 0, 0
    best_avg_ret = -np.inf
    for epoch in range(epochs):
        ep_avg_ret, n_eps = 0, 0
        for step in range(max_steps):
            a, v, logp = ac.step(torch.as_tensor(o, dtype=torch.float32))
            next_o, r, d, t, _ = env.step(a)

            buf.store(o, a, v, r, logp)
            ep_ret += r
            ep_len += 1
            ep_avg_ret += r

            o = next_o

            timeout = ep_len == max_ep_len
            terminal = d or t or timeout
            epoch_ended = step == max_steps - 1

            if terminal or epoch_ended:
                if epoch_ended and not(terminal):
                    print(f"Trajectory cut off by epoch at {ep_len} steps.", flush=True)
                if timeout or epoch_ended:
                    _, v, _ = ac.step(torch.as_tensor(o, dtype=torch.float32))
                else:
                    v = 0
                buf.finish_path(v)
                n_eps += 1
                o, ep_ret, ep_len = env.reset()[0], 0, 0

        loss_pi, loss_v = update()

        ep_avg_ret = ep_avg_ret / n_eps
        if ep_avg_ret > best_avg_ret:
            best_avg_ret = ep_avg_ret
            torch.save(ac.state_dict(), save_path)
        logger.info(
            f"Epoch: {epoch}, Loss pi: {loss_pi}, Loss v: {loss_v}, Rew: {ep_avg_ret}"
        )


def ppo_test(env_fn, actor_critic, ac_kwargs, test_epochs, model_path):
    env = env_fn()
    ac = actor_critic(env.observation_space, env.action_space, **ac_kwargs)
    ac.load_state_dict(torch.load(model_path))
    ac.eval()

    for epoch in range(test_epochs):
        o, ep_ret, ep_len = env.reset()[0], 0, 0
        while True:
            a, v, logp = ac.step(torch.as_tensor(o, dtype=torch.float32))
            next_o, r, d, tructated, _ = env.step(a)
            ep_ret += r
            ep_len += 1
            o = next_o
            terminal = d or tructated
            if terminal:
                print(f"Epoch: {epoch}, Reward: {ep_ret}")
                o, ep_ret, ep_len = env.reset()[0], 0, 0
                break
        cmd = input("Continue? [y/n]")
        if cmd == "n":
            break


if __name__ == "__main__":
    import argparse

    parser = argparse.ArgumentParser()
    parser.add_argument("--test", action="store_true")
    args = parser.parse_args()
    env_name = "LunarLander-v2"
    env_name = "CartPole-v1"
    if args.test:
        ppo_test(
            lambda: gym.make(env_name, render_mode="human"),
            actor_critic=core.MLPActorCritic,
            ac_kwargs=dict(hidden_sizes=[64, 64], activation=nn.Tanh),
            test_epochs=10,
            model_path="ppo.pt",
        )
    else:
        ppo(
            lambda: gym.make(env_name),
            ac_constructor=core.MLPActorCritic,
            ac_kwargs=dict(hidden_sizes=[64, 64], activation=nn.Tanh),
            epochs=500,
            max_steps=5000,
            max_ep_len=1000,
            gamma=0.99,
            lam=0.97,
            clip_ratio=0.2,
            pi_lr=3e-4,
            vf_lr=1e-3,
            train_pi_iters=80,
            train_v_iters=80,
            target_kl=0.01,
            logger=logging.getLogger(__name__),
            save_path="ppo.pt",
        )



import torch
import gymnasium as gym
from typing import List
import cv2
import numpy as np


def get_policy_network(
    sizes: List[int],
    activation: torch.nn.Module = torch.nn.ReLU,
    output_activation: torch.nn.Module = torch.nn.Identity,
):
    layers = []
    for j in range(len(sizes) - 1):
        act = activation if j < len(sizes) - 2 else output_activation
        layers += [torch.nn.Linear(sizes[j], sizes[j + 1]), act()]
    return torch.nn.Sequential(*layers)


def get_policy(obs: torch.Tensor, policy_network: torch.nn.Module):
    logits = policy_network(obs)
    return torch.distributions.Categorical(logits=logits)


def sample_action(obs: torch.Tensor, policy_network: torch.nn.Module):
    return get_policy(obs, policy_network).sample().item()


def compute_loss(
    obs: torch.Tensor,
    act: torch.Tensor,
    weights: torch.Tensor,
    policy_network: torch.nn.Module,
):
    logp = get_policy(obs, policy_network).log_prob(act)
    return -(logp * weights).mean()


env = gym.make("CartPole-v1", render_mode="rgb_array")
obs_dim = env.observation_space.shape[0]
n_acts = env.action_space.n
policy_network = get_policy_network(sizes=[obs_dim, 64, 32, n_acts])
policy_network.train()
optimizer = torch.optim.Adam(policy_network.parameters(), lr=1e-2)
batch_size = 5000
epochs = 100


def train_one_epoch():
    batch_obs = []
    batch_acts = []
    batch_weights = []
    batch_rewards = []
    batch_lens = []

    ep_rewards = []
    obs = env.reset()[0]

    finish_rendering_this_epoch = False

    while True:
        batch_obs.append(obs.copy())
        act = sample_action(torch.as_tensor(obs, dtype=torch.float32), policy_network)
        if not finish_rendering_this_epoch:
            frame = env.render()
            cv2.imshow("frame", frame)
            cv2.waitKey(1)
            
        obs, rew, terminated, truncated, _ = env.step(act)

        batch_acts.append(act)
        ep_rewards.append(rew)
        if terminated or truncated:
            ep_sum, ep_len = sum(ep_rewards), len(ep_rewards)
            batch_rewards.append(ep_sum)
            batch_lens.append(ep_len)
            batch_weights += [ep_sum] * ep_len

            finish_rendering_this_epoch = True
            obs, ep_rewards = env.reset()[0], []
            if len(batch_obs) > batch_size:
                break

    optimizer.zero_grad()
    batch_loss = compute_loss(
        obs=torch.as_tensor(np.array(batch_obs), dtype=torch.float32),
        act=torch.as_tensor(np.array(batch_acts), dtype=torch.int32),
        weights=torch.as_tensor(np.array(batch_weights), dtype=torch.float32),
        policy_network=policy_network,
    )
    batch_loss.backward()
    optimizer.step()
    return batch_loss, batch_rewards, batch_lens

best_reward = 0
for i in range(epochs):
    batch_loss, batch_rewards, batch_lens = train_one_epoch()
    print(
        f"epoch: {i:3d} loss: {batch_loss:.3f} reward: {np.mean(batch_rewards):.3f} ep_len: {np.mean(batch_lens):3f}"
    )
    if np.mean(batch_rewards) > best_reward:
        best_reward = np.mean(batch_rewards)
        torch.save(policy_network.state_dict(), "policy_network.pt")


import torch
import numpy as np
import core
from copy import deepcopy
import gymnasium as gym


class ReplayBuffer(object):
    def __init__(self, obs_dim, act_dim, size) -> None:
        self.obs_buf = np.zeros(core.combined_shape(size, obs_dim), dtype=np.float32)
        self.obs2_buf = np.zeros(core.combined_shape(size, obs_dim), dtype=np.float32)
        self.act_buf = np.zeros(core.combined_shape(size, act_dim), dtype=np.float32)
        self.rew_buf = np.zeros(size, dtype=np.float32)
        self.done_buf = np.zeros(size, dtype=np.float32)
        self.ptr, self.size, self.max_size = 0, 0, size

    def store(self, obs, act, rew, next_obs, done):
        self.obs_buf[self.ptr] = obs
        self.obs2_buf[self.ptr] = next_obs
        self.act_buf[self.ptr] = act
        self.rew_buf[self.ptr] = rew
        self.done_buf[self.ptr] = done

        self.ptr = (self.ptr + 1) % self.max_size
        self.size = min(self.size + 1, self.max_size)

    def sample_batch(self, batch_size=32):
        idxs = np.random.randint(0, self.size, size=batch_size)
        batch = dict(
            obs=self.obs_buf[idxs],
            obs2=self.obs2_buf[idxs],
            act=self.act_buf[idxs],
            rew=self.rew_buf[idxs],
            done=self.done_buf[idxs],
        )
        return {k: torch.as_tensor(v) for k, v in batch.items()}


def ddpg(
    env_fn,
    ac_constructor,
    ac_kwargs,
    epochs,
    steps_per_epoch,
    replay_size=int(1e6),
    gamma=0.99,
    polyak=0.995,
    pi_lr=1e-3,
    q_lr=1e-3,
    batch_size=100,
    start_steps=10000,
    update_after=1000,
    update_every=50,
    act_noise=0.1,
    num_test_episodes=10,
    max_ep_len=1000,
):
    env, test_env = env_fn(), env_fn()
    obs_dim = env.observation_space.shape
    act_dim = env.action_space.shape[0]

    act_limit = env.action_shape.high[0]
    ac = ac_constructor(env.observation_space, env.action_space, **ac_kwargs)
    ac_targ = deepcopy(ac)

    replay_buffer = ReplayBuffer(obs_dim=obs_dim, act_dim=act_dim, size=replay_size)

    for p in ac_targ.parameters():
        p.requres_grad = False

    def compute_loss_q(data):
        o, a, r, o2, d = (
            data["obs"],
            data["act"],
            data["rew"],
            data["obs2"],
            data["done"],
        )
        q = ac.q(o, a)

        with torch.no_grad():
            q_pi_targ = ac_targ.q(o2, ac_targ.pi(o2))
            backup = r + gamma * (1 - d) * q_pi_targ

        loss_q = ((q - backup) ** 2).mean()
        loss_info = dict(QVals=q.detach().numpy())
        return loss_q, loss_info

    def compute_loss_pi(data):
        o = data["obs"]
        q_pi = ac.q(o, ac.pi(o))
        return -q_pi.mean()

    pi_optimizer = torch.optim.Adam(ac.pi.parameters(), lr=pi_lr)
    q_optimizer = torch.optim.Adam(ac.q.parameters(), lr=q_lr)

    def update(data):
        q_optimizer.zero_grad()
        loss_q, loss_info = compute_loss_q(data)
        loss_q.backward()
        q_optimizer.step()

        for p in ac.q.parameters():
            p.requires_grad = False

        pi_optimizer.zero_grad()
        loss_pi = compute_loss_pi(data)
        loss_pi.backward()
        pi_optimizer.step()

        for p in ac.q.parameters():
            p.requires_grad = True

        with torch.no_grad():
            for p, p_targ in zip(ac.parameters(), ac_targ.parameters()):
                p_targ.data.mul_(polyak)
                p_targ.data.add_((1 - polyak) * p.data)

    def get_action(o, noise_scale):
        a = ac.act(torch.as_tensor(o, dtype=torch.float32))
        a += noise_scale * np.random.randn(act_dim)
        return np.clip(a, -act_limit, act_limit)

    def test_agent():
        for _ in range(num_test_episodes):
            o, d, ep_ret, ep_len = test_env.reset()[0], False, 0, 0
            while not (d or (ep_len == max_ep_len)):
                o, r, terminal, trunctated, _ = test_env.step(get_action(o, 0))
                d = terminal or trunctated
            ep_ret += r
            ep_len += 1
        return ep_ret, ep_len

    total_steps = steps_per_epoch * epochs
    o, ep_ret, ep_len = env.reset()[0], 0, 0
    best_avg_ret = -np.inf

    for t in range(total_steps):
        if t > start_steps:
            a = get_action(o, act_noise)
        else:
            a = env.action_space.sample()

        (
            o2,
            r,
            terminal,
            trunctated,
            _,
        ) = env.step(a)
        d = terminal or trunctated
        ep_ret += r
        ep_len += 1

        d = False if ep_len == max_ep_len else d

        replay_buffer.store(o, a, r, o2, d)

        o = o2
        if d or (ep_len == max_ep_len):
            o, ep_ret, ep_len = env.reset()[0], 0, 0

        if t >= update_after and t % update_every == 0:
            for _ in range(update_every):
                batch = replay_buffer.sample_batch(batch_size)
                update(data=batch)

        if (t + 1) % steps_per_epoch == 0:
            epoch = (t + 1) // steps_per_epoch
            ret, len = test_agent()
            avg_ep_ret = ret / len
            if avg_ep_ret > best_avg_ret:
                best_avg_ret = avg_ep_ret
                torch.save(ac.state_dict(), "ddpg.pt")
            print(f"Epoch: {epoch}, Average ep ret: {avg_ep_ret}")


if __name__ == "__main__":
    ddpg(
        lambda: gym.make("CarPole-v1"),
        ac_constructor=core.MLPQActorCritic,
        ac_kwargs=dict(hidden_sizes=[256, 256]),
        epochs=50,
    )
